@@ -11,6 +11,7 @@ import platform
 from binascii import hexlify, unhexlify
 from .filetimes import filetime_from_dec
 from datetime import datetime, timedelta
+from binascii import b2a_qp, a2b_qp
 import sys
 import time
 from functools import wraps
@@ -40,13 +41,17 @@ def retry(f, timeout=0.25, num_tries=4):
     return wrapper
 
 
-def _bytes(s):
-    if isinstance(s, unicode):
-        return s.encode('utf-8', 'replace')
-    return bytes(s)
+def _to_hex(s):
+    return unicode(b2a_qp(s), 'ascii').replace('=', '\\x')
 
 
-def gen_key(cont, local=True, silent=False):
+def _from_hex(s):
+    if isinstance(s, bytes):
+        s = unicode(s, 'ascii')
+    return a2b_qp(s.replace('\\x', '='))
+
+
+def gen_key(cont, local=True, silent=False, provider=None):
     '''
     Создание контейнера и двух пар ключей в нем
 
@@ -55,18 +60,21 @@ def gen_key(cont, local=True, silent=False):
     :silent: Если True, включает режим без диалоговых окон. Без аппаратного датчика случайных
         чисел в таком режиме контейнер создать невозможно!
         По умолчанию silent=False
+    :provider: Если не None, флаг local игнорируется и криптопровайдер
+        выбирается принудительно
     :returns: True, если операция успешна
 
     '''
     silent_flag = csp.CRYPT_SILENT if silent else 0
-    provider = str("Crypto-Pro HSM CSP") if not local else None
-    cont = _bytes(cont)
+    if provider is None:
+        provider = str("Crypto-Pro HSM CSP") if not local else None
+    cont = _from_hex(cont)
 
     try:
         ctx = csp.Crypt(cont, csp.PROV_GOST_2001_DH, silent_flag, provider)
     except (ValueError, SystemError):
 
-        if platform.system() == 'Linux' and local:
+        if platform.system() == 'Linux' and provider is None:
             cont = b'\\\\.\\HDIMAGE\\' + cont
 
         ctx = csp.Crypt(cont, csp.PROV_GOST_2001_DH, csp.CRYPT_NEWKEYSET |
@@ -90,22 +98,25 @@ def gen_key(cont, local=True, silent=False):
     return True
 
 
-def remove_key(cont, local=True):
+def remove_key(cont, local=True, provider=None):
     '''
     Удаление контейнера
 
     :cont: Имя контейнера
-    :local: Если True, контейнер создается в локальном хранилище по умолчанию
+    :local: Если True, контейнер удаляется в локальном хранилище по умолчанию
+    :provider: Если не None, флаг local игнорируется и криптопровайдер
+        выбирается принудительно
     :returns: True, если операция успешна
 
     '''
-    cont = _bytes(cont)
-    provider = str("Crypto-Pro HSM CSP") if not local else None
+    cont = _from_hex(cont)
+    if provider is None:
+        provider = str("Crypto-Pro HSM CSP") if not local else None
     csp.Crypt.remove(cont, csp.PROV_GOST_2001_DH, provider)
     return True
 
 
-def create_request(cont, params, local=True):
+def create_request(cont, params, local=True, provider=None):
     """Создание запроса на сертификат
 
     :cont: Имя контейнера
@@ -135,12 +146,15 @@ def create_request(cont, params, local=True):
             закодированных в DER-кодировку внешними средставми
         }
     :local: Если True, работа идет с локальным хранилищем
+    :provider: Если не None, флаг local игнорируется и криптопровайдер
+        выбирается принудительно
     :returns: байтовая строка с запросом в DER-кодировке
 
     """
 
-    provider = str("Crypto-Pro HSM CSP") if not local else None
-    cont = _bytes(cont)
+    if provider is None:
+        provider = str("Crypto-Pro HSM CSP") if not local else None
+    cont = _from_hex(cont)
     ctx = csp.Crypt(cont, csp.PROV_GOST_2001_DH, 0, provider)
     req = csp.CertRequest(ctx, )
     set_q_defaults(params)
@@ -164,18 +178,21 @@ def create_request(cont, params, local=True):
     return req.get_data()
 
 
-def bind_cert_to_key(cont, cert, local=True):
+def bind_cert_to_key(cont, cert, local=True, provider=None):
     """Привязка сертификата к закрытому ключу в контейнере
 
     :cont: Имя контейнера
     :cert: Сертификат в байтовой строке
     :local: Если True, работа идет с локальным хранилищем
+    :provider: Если не None, флаг local игнорируется и криптопровайдер
+        выбирается принудительно
     :returns: отпечаток сертификата в виде строки
 
     """
-    provider = str("Crypto-Pro HSM CSP") if not local else None
+    if provider is None:
+        provider = str("Crypto-Pro HSM CSP") if not local else None
     cert = autopem(cert)
-    cont = _bytes(cont)
+    cont = _from_hex(cont)
     ctx = csp.Crypt(cont, csp.PROV_GOST_2001_DH, 0, provider)
     newc = csp.Cert(cert)
     newc.bind(ctx)
@@ -308,12 +325,10 @@ def decrypt(data, thumb):
     cs = csp.CertStore(None, b"MY")
     certs = list(cs.find_by_thumb(unhexlify(thumb)))
     assert len(certs), 'Certificate for thumbprint not found'
-    decrcs = csp.CertStore()
-    decrcs.add_cert(certs[0])
     bin_data = data
     msg = csp.CryptMsg(bin_data)
-    decrypted = msg.decrypt(decrcs)
-    return decrypted
+    msg.decrypt_by_cert(certs[0])
+    return msg.get_data()
 
 
 def pkcs7_info(data):
